<?php

namespace Drupal\currencyapi;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\KeyValueStore\KeyValueExpirableFactoryInterface;

/**
 * Provides currency exchange rates from CurrencyAPI.com.
 */
class CurrencyApiService {

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The key-value expirable storage.
   *
   * @var \Drupal\Core\KeyValueStore\KeyValueExpirableFactoryInterface
   */
  protected $keyValueExpirable;

  /**
   * Constructs a new CurrencyApiService.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    KeyValueExpirableFactoryInterface $key_value_expirable
  ) {
    $this->configFactory = $config_factory;
    $this->keyValueExpirable = $key_value_expirable;
  }

  /**
   * Fetches exchange rates from CurrencyAPI.com.
   *
   * @return array|null
   *   The exchange rates data or NULL on failure.
   */
  public function fetchExchangeRates(): ?array {

    $d = [];

    $d['config'] = $this->configFactory->get('currencyapi.settings');
    $d['api_key'] = $d['config']->get('api_key');
    $d['base_currency'] = $d['config']->get('base_currency') ?? 'EUR';
    $d['target_currencies'] = $d['config']->get('target_currencies') ?? ['USD'];
    $d['currencies'] = explode(",",$d['config']->get('currencies'));
    $d['default_currency'] = $d['currencies'][0];

    if (empty($d['api_key'])) {
      throw new \RuntimeException('CurrencyAPI.com API key is not configured.');
    }

    $d['url'] = $this->buildApiUrl($d['api_key'], $d['currencies']);

    // Live query
    $d['response'] = $this->makeApiRequest($d['url']);

    // Fake query
    // $d['fake_response'] = [
    //   "meta" => [],
    //   "data" =>  [
    //     "CHF" => [
    //       "code" => "CHF",
    //       "value" => 0.9379563304,
    //     ],
    //     "EUR" => [
    //       "code" => "EUR",
    //       "value" => 1,
    //     ],
    //     "GBP" => [
    //       "code" => "GBP",
    //       "value" => 0.8390516042,
    //     ],
    //     "JPY" => [
    //       "code" => "JPY",
    //       "value" => 163.4619011136,
    //     ],
    //     "USD" => [
    //       "code" => "USD",
    //       "value" => 1.1338380637,
    //     ],
    //   ],
    // ];
    //
    // $d['response'] = $d['fake_response'];

    $this->setDefaultCurrency($d);

    if ($d['response'] && isset($d['response']['data'])) {
      $this->storeRates($d['response']['data'], $d['default_currency']);
      return $d['response'];
    }
    return NULL;
  }


  protected function setDefaultCurrency(&$d) {
    foreach ($d['response']['data'] as $code => $value) {
      $d['response']['data'][$code]['is_default'] = FALSE;
      if ($code == $d['default_currency']) {
        $d['response']['data'][$code]['is_default'] = TRUE;
      }
    }
  }

  /**
   * Builds the API URL.
   */
  protected function buildApiUrl(string $api_key, array $currencies): string {

    return sprintf(
      'https://api.currencyapi.com/v3/latest?apikey=%s&base_currency=%s&currencies=%s',
      urlencode($api_key),
      urlencode($currencies[0]),
      urlencode(implode(',', $currencies))
    );
  }

  /**
   * Makes the API request using file_get_contents().
   */
  protected function makeApiRequest(string $url): ?array {

    $d = [];

    $d['url'] = $url;

    $d['context'] = stream_context_create([
      'http' => [
        'timeout' => 10,
        'header' => "Accept: application/json\r\n"
      ]
    ]);

    $d['response'] = @file_get_contents($d['url'], false, $d['context']);

    if ($d['response'] === FALSE) {
      \Drupal::logger('currencyapi')->error('Failed to fetch exchange rates from CurrencyAPI.com');
      return NULL;
    }

    return json_decode($d['response'], TRUE);
  }

  /**
   * Stores exchange rates in key-value storage.
   */
  protected function storeRates(array $rates, string $default_currency): void {
    $d = [];
    $d['rates'] = $rates;
    $d['default_currency'] = $default_currency;
    $d['storage'] = $this->keyValueExpirable->get('currencyapi_rates');
    $d['now'] =  time() + 3600; //$this->time->getRequestTime();
    $d['expiration'] = ($this->configFactory->get('currencyapi.settings')->get('cache_expiration') ?? 86400);
    $d['storage']->setWithExpire("currencyapi", $d['rates'],  $d['expiration']);
  }

  /**
   * Gets the exchange rate for a specific currency.
   */
  public function getRate(string $currency): ?float {}

  /**
   * Gets all stored exchange rates.
   */
  public function getAllRates(): array {
    $d = [];
    $d['storage'] = $this->keyValueExpirable->get('currencyapi_rates');
    $d['all'] = $d['storage']->getAll();
    return $d['all'];
  }

  /**
   * Converts an amount between currencies.
   */
  public function convert(float $amount, string $from, string $to) {

    $d = [];
    $d['amount'] = $amount;
    $d['from'] = $from;
    $d['to'] = $to;

    if ($d['from'] === $d['to']) {
      $d['output'] = $d['amount'];
    }

    $d['rates'] = $this->getAllRates();

    // Direct conversion available.
    // @todo replace EUR with store default currency
    if ($d['from'] === 'EUR' && isset($d['rates']['currencyapi'][$d['to']])) {
      $d['output'] = ($d['amount'] * $d['rates']['currencyapi'][$d['to']]['value']);
      $d['output'] = number_format($d['output'], 2, ',', ' ');
    }

    return $d['output'];

    // @todo Conversion through min currency
    // if (
    //    //isset($d['rates'][$d['from']])
    //    //&& isset($d['rates']['currencyapi']['code'])
    //   $d['rates']['currencyapi']['code'] == $d['to']
    //   ) {
    //   $d['amount_in_destination_currency'] = $d['amount'] * $d['rates']['currencyapi']['value'];
    //   $d['output'] = $d['amount_in_destination_currency'];
    // }
    // dump ($d);
    // exit;

  }

  /**
   * Gets the last update timestamp.
   */
  public function getLastUpdate(): ?int {
    $storage = $this->keyValueExpirable->get('currencyapi_rates');
    return $storage->get('_last_update');
  }


  public function getSystemCurrencies()  {}

}
